package org.commonjava.tensor.flat.data.store;

import static org.apache.commons.io.FileUtils.readFileToString;
import static org.apache.commons.io.FileUtils.write;
import static org.commonjava.tensor.flat.util.FlatFilenameFormatter.getFilename;

import java.io.File;
import java.io.IOException;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;

import org.commonjava.tensor.data.TensorDataException;
import org.commonjava.tensor.data.store.IndexStore;
import org.commonjava.tensor.flat.config.FlatFileTensorConfiguration;
import org.commonjava.util.logging.Logger;
import org.commonjava.web.json.ser.JsonSerializer;

import com.google.common.cache.Cache;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheBuilderSpec;
import com.google.common.cache.RemovalListener;
import com.google.common.cache.RemovalNotification;
import com.google.gson.reflect.TypeToken;

public class FlatFileIndexStore<K, V>
    implements IndexStore<K, V>, RemovalListener<K, V>, Callable<V>
{

    // FIXME: Total hack!
    // 
    // To work around the losing real generic reference through multiple
    // constructors when passing in TypeToken<V> to an embedded FileLoader<K,V>
    // class, we're making this class itself a Callable<V>. This means we need
    // the reference to the key to load when call() is invoked.
    private final ThreadLocal<K> loadableKey = new ThreadLocal<K>();

    private final Logger logger = new Logger( getClass() );

    private static final long DEFAULT_MAX_CACHE_SIZE = 1000;

    protected File storageDir;

    protected final JsonSerializer serializer;

    protected Cache<K, V> cache;

    protected final String name;

    protected Class<V> type;

    private final TypeToken<V> typeToken;

    public FlatFileIndexStore( final String name, final Class<V> type, final JsonSerializer serializer,
                               final FlatFileTensorConfiguration config )
    {
        this.name = name;
        this.type = type;
        this.typeToken = null;
        this.serializer = serializer;

        init( config );
    }

    protected void init( final FlatFileTensorConfiguration config )
    {
        this.storageDir = new File( config.getStorageBasedir(), name );
        storageDir.mkdirs();

        logger.info( "[%s] Storing to: %s", name, storageDir );

        String cacheConfig = config.getCacheConfig( name );
        if ( cacheConfig == null )
        {
            cacheConfig = config.getDefaultCacheConfig();
        }

        if ( cacheConfig != null )
        {
            final CacheBuilderSpec spec = CacheBuilderSpec.parse( cacheConfig );
            this.cache = CacheBuilder.from( spec )
                                     .removalListener( this )
                                     .build();
        }
        else
        {
            this.cache = CacheBuilder.newBuilder()
                                     .weakKeys()
                                     .weakValues()
                                     .maximumSize( DEFAULT_MAX_CACHE_SIZE )
                                     .build();
        }
    }

    public FlatFileIndexStore( final String name, final TypeToken<V> typeToken, final JsonSerializer serializer,
                               final FlatFileTensorConfiguration config )
    {
        this.name = name;
        this.typeToken = typeToken;
        this.type = null;
        this.serializer = serializer;

        init( config );
    }

    File getStorageDir()
    {
        return storageDir;
    }

    @Override
    public final V store( final K key, final V value )
        throws TensorDataException
    {
        V old = null;
        try
        {
            if ( contains( key ) )
            {
                loadableKey.set( key );
                old = cache.get( key, this );
            }
        }
        catch ( final ExecutionException e )
        {
            new Logger( getClass() ).error( "[%s] Failed to read original value: %s. Reason: %s", e, name, key,
                                            e.getMessage() );
        }
        finally
        {
            loadableKey.remove();
        }

        final String fname = getFilename( key );
        final File f = new File( storageDir, fname );

        writeValue( f, key, value );

        cache.put( key, value );

        return old;
    }

    @Override
    public final V get( final K key )
        throws TensorDataException
    {
        if ( key == null )
        {
            return null;
        }

        if ( !contains( key ) )
        {
            return null;
        }

        try
        {
            loadableKey.set( key );
            return cache.get( key, this );
        }
        catch ( final ExecutionException e )
        {
            throw new TensorDataException( "[%s] Failed to retrieve: %s from disk. Reason: %s", e, name, key,
                                           e.getMessage() );
        }
        finally
        {
            loadableKey.remove();
        }
    }

    @Override
    public final void clear()
    {
        //        final String[] listing = storageDir.list();
        //        for ( final String fname : listing )
        //        {
        //            final File f = new File( storageDir, fname );
        //            f.delete();
        //        }
        cache.invalidateAll();
    }

    @Override
    public final boolean contains( final K key )
    {
        final String fname = getFilename( key );
        final File f = new File( storageDir, fname );
        return f.exists();
    }

    @Override
    public final V call()
        throws Exception
    {
        final K key = loadableKey.get();
        //        logger.info( "[%s] Loading: %s", name, key );

        if ( key == null )
        {
            throw new ExecutionException( "[" + name + "] No key found for loading from cache.",
                                          new IllegalStateException() );
        }

        final String fname = getFilename( key );
        final File f = new File( storageDir, fname );
        if ( f.exists() )
        {
            return readValue( f );
        }

        return null;
    }

    protected void writeValue( final File f, final K key, final V value )
        throws TensorDataException
    {
        //        logger.info( "[%s] Storing: %s = %s", name, key, value );
        final String json = serializer.toString( value );

        try
        {
            write( f, json );
        }
        catch ( final IOException e )
        {
            throw new TensorDataException( "[%s] Failed to write: %s. Reason: %s", e, name, f, e.getMessage() );
        }
    }

    protected V readValue( final File f )
        throws TensorDataException
    {
        try
        {
            final String json = readFileToString( f );
            V result = null;
            if ( type != null )
            {
                result = serializer.fromString( json, type );
            }
            else if ( typeToken != null )
            {
                result = serializer.fromString( json, typeToken );
            }

            return result;
        }
        catch ( final IOException e )
        {
            throw new TensorDataException( "[%s] Failed to read: %s. Reason: %s", e, name, f, e.getMessage() );
        }
    }

    @Override
    public void onRemoval( final RemovalNotification<K, V> event )
    {
        final K key = event.getKey();
        //        logger.info( "[%s] Removing: %s", name, key );

        final String fname = getFilename( key );
        final File f = new File( storageDir, fname );
        if ( f.exists() )
        {
            f.delete();
        }
    }

    @Override
    public V remove( final K key )
        throws TensorDataException
    {
        if ( contains( key ) )
        {
            final V result = get( key );
            cache.invalidate( key );

            return result;
        }

        return null;
    }

}
