package org.commonjava.tensor.flat.data.store;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.io.FileUtils;
import org.apache.maven.graph.effective.ref.EProjectKey;
import org.apache.maven.graph.effective.rel.ProjectRelationship;
import org.commonjava.tensor.data.TensorDataException;
import org.commonjava.tensor.data.store.IndexStore;
import org.commonjava.tensor.flat.config.FlatFileTensorConfiguration;
import org.commonjava.util.logging.Logger;
import org.commonjava.web.json.ser.JsonSerializer;

public class FlatFileIndexStoreWithMetadata<V>
{

    private static final String KEYS_FILENAME = "keys.json";

    private final Logger logger = new Logger( getClass() );

    private final String name;

    private final Map<EProjectKey, EProjectKey> keys = new ConcurrentHashMap<EProjectKey, EProjectKey>();

    private final IndexStore<EProjectKey, V> store;

    private final File storage;

    private final JsonSerializer serializer;

    private final Class<V> valueType;

    public FlatFileIndexStoreWithMetadata( final String name, final Class<V> valueType,
                                           final IndexStore<EProjectKey, V> store, final JsonSerializer serializer,
                                           final FlatFileTensorConfiguration config )
        throws TensorDataException
    {
        this.valueType = valueType;
        this.serializer = serializer;

        final File d = config.getStorageBasedir();
        storage = new File( d, name + "-" + KEYS_FILENAME );
        storage.getParentFile()
               .mkdirs();

        logger.info( "[%s] Storing relationships of type %s to: %s", name, valueType.getName(), storage );

        this.name = name;
        this.store = store;

        loadKeys();
    }

    public boolean contains( final EProjectKey key )
    {
        return keys.containsKey( key );
    }

    public V store( final EProjectKey key, final V value )
        throws TensorDataException
    {
        final V result = store.store( key, value );

        if ( keys.put( key, key ) == null )
        {
            appendKey( key );
        }

        return result;
    }

    public V remove( final EProjectKey key )
        throws TensorDataException
    {
        final V result = store.remove( key );
        if ( result != null )
        {
            writeKeys();
        }

        return result;
    }

    private void loadKeys()
        throws TensorDataException
    {
        if ( storage.exists() )
        {
            try
            {
                final List<String> lines = FileUtils.readLines( storage );
                for ( final String line : lines )
                {
                    if ( line.trim()
                             .startsWith( "#" ) )
                    {
                        continue;
                    }

                    final EProjectKey key = serializer.fromString( line, EProjectKey.class );
                    keys.put( key, key );
                }
            }
            catch ( final IOException e )
            {
                throw new TensorDataException( "[%s] Failed to read keys from file: %s. Reason: %s", e, name, storage,
                                               e.getMessage() );
            }
        }
    }

    private void writeKeys()
        throws TensorDataException
    {
        logger.info( "[%s] Writing keys to: %s", name, storage );

        final Set<String> json = new HashSet<String>();
        json.add( String.format( "# Keys for %s", name ) );

        for ( final EProjectKey key : keys.keySet() )
        {
            json.add( serializer.toString( key ) );
        }

        try
        {
            FileUtils.writeLines( storage, json );
        }
        catch ( final IOException e )
        {
            throw new TensorDataException( "[%s] Failed to write key enumeration file: %s. Reason: %s", e, name,
                                           storage, e.getMessage() );
        }
    }

    private void appendKey( final EProjectKey key )
        throws TensorDataException
    {
        logger.info( "[%s] Appending key: %s to: %s", name, key, storage );
        try
        {
            final String json = serializer.toString( key );
            FileUtils.writeLines( storage, Collections.singleton( json ), true );
        }
        catch ( final IOException e )
        {
            throw new TensorDataException( "[%s] Failed to append %s to key enumeration file: %s. Reason: %s", e, name,
                                           key, storage, e.getMessage() );
        }
    }

    public V get( final EProjectKey key )
        throws TensorDataException
    {
        return store.get( key );
    }

    @SuppressWarnings( "rawtypes" )
    private V constructResult( final EProjectKey key, final Set<ProjectRelationship> rels )
        throws TensorDataException
    {
        if ( rels == null )
        {
            return null;
        }

        try
        {
            return valueType.getConstructor( EProjectKey.class, Collection.class )
                            .newInstance( key, rels );
        }
        catch ( final InstantiationException e )
        {
            throw new TensorDataException(
                                           "[%s] Failed to construct new value of type: %s using key and relationship collection. Reason: %s",
                                           e, name, valueType.getName(), e.getMessage() );
        }
        catch ( final IllegalAccessException e )
        {
            throw new TensorDataException(
                                           "[%s] Failed to construct new value of type: %s using key and relationship collection. Reason: %s",
                                           e, name, valueType.getName(), e.getMessage() );
        }
        catch ( final InvocationTargetException e )
        {
            throw new TensorDataException(
                                           "[%s] Failed to construct new value of type: %s using key and relationship collection. Reason: %s",
                                           e, name, valueType.getName(), e.getTargetException()
                                                                          .getMessage() );
        }
        catch ( final NoSuchMethodException e )
        {
            throw new TensorDataException(
                                           "[%s] Failed to construct new value of type: %s using key and relationship collection. Reason: %s",
                                           e, name, valueType.getName(), e.getMessage() );
        }
    }

    public Set<EProjectKey> keys()
    {
        return Collections.unmodifiableSet( keys.keySet() );
    }

    public int size()
    {
        return keys.size();
    }

}
