package org.commonjava.tensor.flat.data.store;

import static org.apache.commons.io.FileUtils.readFileToString;
import static org.apache.commons.io.FileUtils.write;
import static org.commonjava.tensor.flat.util.FlatFilenameFormatter.getFilename;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.io.FileUtils;
import org.commonjava.tensor.data.TensorDataException;
import org.commonjava.tensor.flat.config.FlatFileTensorConfiguration;
import org.commonjava.util.logging.Logger;
import org.commonjava.web.json.ser.JsonSerializer;

import com.google.common.cache.RemovalNotification;

public class MultiValueFlatFileIndexStore<K, C>
    extends FlatFileIndexStore<K, Set<C>>
{

    private final Logger logger = new Logger( getClass() );

    private final Class<C> constituentType;

    public MultiValueFlatFileIndexStore( final String name, final Class<C> constituentType,
                                         final JsonSerializer serializer, final FlatFileTensorConfiguration config )
    {
        super( name, (Class<Set<C>>) null, serializer, config );

        this.constituentType = constituentType;
    }

    @Override
    protected synchronized void writeValue( final File d, final K key, final Set<C> value )
        throws TensorDataException
    {
        final File dir = new File( d.getParentFile(), d.getName() + ".new" );
        dir.mkdirs();

        File f = null;
        try
        {
            for ( final C item : value )
            {
                final String fname = getFilename( item );

                f = new File( dir, fname );

                //        logger.info( "[%s] Storing: %s = %s", name, key, value );
                final String json = serializer.toString( item );

                write( f, json );
            }

            if ( d.exists() )
            {
                FileUtils.forceDelete( d );
            }

            dir.renameTo( d );
        }
        catch ( final IOException e )
        {
            try
            {
                FileUtils.forceDelete( dir );
            }
            catch ( final IOException eInner )
            {
                logger.error( "[%s] Failed to delete aborted value-directory for: %s after write failed. Reason: %s",
                              eInner, name, key, eInner.getMessage() );
            }

            throw new TensorDataException( "[%s] Failed to write: %s. Reason: %s", e, name, f, e.getMessage() );
        }
    }

    @Override
    protected synchronized Set<C> readValue( final File d )
        throws TensorDataException
    {
        final String[] listing = d.list();
        final List<C> values = new ArrayList<C>();
        for ( final String file : listing )
        {
            final File f = new File( d, file );
            if ( f.isDirectory() )
            {
                // support metadata directories embedded in this directory.
                continue;
            }

            String json;
            try
            {
                json = readFileToString( f );
            }
            catch ( final IOException e )
            {
                throw new TensorDataException( "[%s] Failed to read: %s. Reason: %s", e, name, f, e.getMessage() );
            }

            final C result = serializer.fromString( json, constituentType );

            if ( result != null )
            {
                values.add( constituentType.cast( result ) );
            }
        }

        return new HashSet<C>( values );
    }

    @Override
    public void onRemoval( final RemovalNotification<K, Set<C>> event )
    {
        final K key = event.getKey();
        //        logger.info( "[%s] Removing: %s", name, key );

        final String fname = getFilename( key );
        final File f = new File( storageDir, fname );
        if ( f.exists() )
        {
            try
            {
                FileUtils.forceDelete( f );
            }
            catch ( final IOException e )
            {
                logger.error( "[%s] Failed to remove multi-valued directory for: %s. Reason: %s", e, name, key,
                              e.getMessage() );
            }
        }
    }

}
